@@ -163,6 +163,9 @@ typedef long long int64_t;
 #define METHOD_GET 1
 #define METHOD_HEAD 2
 #define METHOD_POST 3
+#define METHOD_PUT 4
+#define METHOD_DELETE 5
+#define METHOD_TRACE 6
 
 
 /* A multi-family sockaddr. */
@@ -244,7 +247,8 @@ static void read_config( char* filename );
 static void value_required( char* name, char* value );
 static void no_value_required( char* name, char* value );
 static int initialize_listen_socket( usockaddr* usaP );
-static void handle_request( void );
+static void handle_request( void ) __attribute__((noreturn));
+static void finish_request( int exitstatus ) __attribute__((noreturn));
 static void de_dotdot( char* f );
 static int get_pathinfo( void );
 static void do_file( void );
@@ -269,7 +273,7 @@ static int send_error_file( char* filename );
 static void send_error_tail( void );
 static void add_headers( int s, char* title, char* extra_header, char* me, char* mt, off_t b, time_t mod );
 static void start_request( void );
-static void add_to_request( char* str );
+static void add_to_request( char* str, size_t len );
 static char* get_request_line( void );
 static void start_response( void );
 static void add_to_response( char* str );
@@ -279,9 +283,10 @@ static void send_via_sendfile( int fd, int s, off_t size );
 static ssize_t my_read( char* buf, size_t size );
 static ssize_t my_write( void* buf, size_t size );
 #ifdef HAVE_SENDFILE
-static int my_sendfile( int fd, int s, off_t offset, size_t nbytes );
+static ssize_t my_sendfile( int fd, int s, off_t offset, size_t nbytes );
 #endif /* HAVE_SENDFILE */
 static void add_str( char** bufP, size_t* bufsizeP, size_t* buflenP, char* str );
+static void add_data( char** bufP, size_t* bufsizeP, size_t* buflenP, char* str, size_t len );
 static void make_log_entry( void );
 static void check_referrer( void );
 static int really_check_referrer( void );
@@ -553,9 +558,10 @@ main( int argc, char** argv )
 	SSL_load_error_strings();
 	SSLeay_add_ssl_algorithms();
 	ssl_ctx = SSL_CTX_new( SSLv23_server_method() );
-	SSL_CTX_set_options( ssl_ctx, SSL_OP_NO_SSLv3 );
+	SSL_CTX_set_options( ssl_ctx, SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3 );
 	if ( certfile[0] != '\0' )
 	    if ( SSL_CTX_use_certificate_file( ssl_ctx, certfile, SSL_FILETYPE_PEM ) == 0 ||
+		 SSL_CTX_use_certificate_chain_file( ssl_ctx, certfile ) == 0 ||
 		 SSL_CTX_use_PrivateKey_file( ssl_ctx, certfile, SSL_FILETYPE_PEM ) == 0 ||
 		 SSL_CTX_check_private_key( ssl_ctx ) == 0 )
 		{
@@ -857,7 +863,6 @@ main( int argc, char** argv )
 	    if ( listen6_fd != -1 )
 		(void) close( listen6_fd );
 	    handle_request();
-	    exit( 0 );
 	    }
 	(void) close( conn_fd );
 	}
@@ -1196,7 +1201,7 @@ handle_request( void )
 	if ( SSL_accept( ssl ) == 0 )
 	    {
 	    ERR_print_errors_fp( stderr );
-	    exit( 1 );
+	    finish_request( 1 );
 	    }
 	}
 #endif /* USE_SSL */
@@ -1205,15 +1210,14 @@ handle_request( void )
     start_request();
     for (;;)
 	{
-	char buf[10000];
+	char buf[50000];
 	int rr = my_read( buf, sizeof(buf) - 1 );
 	if ( rr < 0 && ( errno == EINTR || errno == EAGAIN ) )
 	    continue;
 	if ( rr <= 0 )
 	    break;
 	(void) alarm( READ_TIMEOUT );
-	buf[rr] = '\0';
-	add_to_request( buf );
+	add_to_request( buf, rr );
 	if ( strstr( request, "\015\012\015\012" ) != (char*) 0 ||
 	     strstr( request, "\012\012" ) != (char*) 0 )
 	    break;
@@ -1273,7 +1277,8 @@ handle_request( void )
 	    cp = &line[5];
 	    cp += strspn( cp, " \t" );
 	    host = cp;
-	    if ( strchr( host, '/' ) != (char*) 0 || host[0] == '.' )
+	    if ( host[0] == '\0' || host[0] == '.' ||
+		 strchr( host, '/' ) != (char*) 0 )
 		send_error( 400, "Bad Request", "", "Can't parse request." );
 	    }
 	else if ( strncasecmp( line, "If-Modified-Since:", 18 ) == 0 )
@@ -1308,6 +1313,12 @@ handle_request( void )
 	method = METHOD_HEAD;
     else if ( strcasecmp( method_str, get_method_str( METHOD_POST ) ) == 0 )
 	method = METHOD_POST;
+    else if ( strcasecmp( method_str, get_method_str( METHOD_PUT ) ) == 0 )
+	method = METHOD_PUT;
+    else if ( strcasecmp( method_str, get_method_str( METHOD_DELETE ) ) == 0 )
+	method = METHOD_DELETE;
+    else if ( strcasecmp( method_str, get_method_str( METHOD_TRACE ) ) == 0 )
+	method = METHOD_TRACE;
     else
 	send_error( 501, "Not Implemented", "", "That method is not implemented." );
 
@@ -1388,6 +1399,63 @@ handle_request( void )
 #ifdef USE_SSL
     SSL_free( ssl );
 #endif /* USE_SSL */
+
+    finish_request( 0 );
+    }
+
+
+static void
+finish_request( int exitstatus )
+    {
+#undef LINGER_SOCKOPT
+#define LINGER_READ
+
+#define LINGER_SECS 5
+
+#ifdef LINGER_SOCKOPT
+    /* The sockopt version of lingering close. Doesn't actually work. */
+    struct linger lin;
+
+    shutdown( conn_fd, SHUT_WR );
+    lin.l_onoff = 1;
+    lin.l_linger = LINGER_SECS;
+    (void) setsockopt(
+	conn_fd, SOL_SOCKET, SO_LINGER, (void*) &lin, sizeof(lin) );
+#endif /* LINGER_SOCKOPT */
+
+#ifdef LINGER_READ
+    /* The "non-blocking read until error/eof/timeout" version of
+    ** lingering close.
+    */
+    int flags;
+    fd_set rfds;
+    struct timeval tv;
+    int r;
+    char* buf[1024];
+	    
+    flags = fcntl( conn_fd, F_GETFL, 0 );
+    if ( flags != -1 )
+	{
+	flags |= (int) O_NDELAY;
+	(void) fcntl( conn_fd, F_SETFL, flags );
+	}
+    shutdown( conn_fd, SHUT_WR );
+    for (;;)
+	{
+	FD_ZERO( &rfds );
+	FD_SET( conn_fd, &rfds );
+	tv.tv_sec = LINGER_SECS;
+	tv.tv_usec = 0;
+	r = select( conn_fd + 1, &rfds, (fd_set*) 0, (fd_set*) 0, &tv );
+	if ( r <= 0 )	/* timeout or error */
+	    break;
+	r = read( conn_fd, (void*) buf, sizeof(buf) );
+	if ( r <= 0 )	/* eof or error */
+	    break;
+	}
+#endif /* LINGER_READ */
+
+    exit( exitstatus );
     }
 
 
@@ -1508,9 +1576,12 @@ do_file( void )
 	do_cgi();
 	return;
 	}
-    else if ( pathinfo != (char*) 0 )
+    if ( pathinfo != (char*) 0 )
 	send_error( 404, "Not Found", "", "File not found." );
 
+    if ( method != METHOD_GET && method != METHOD_HEAD )
+	send_error( 501, "Not Implemented", "", "That method is not implemented." );
+
     fd = open( file, O_RDONLY );
     if ( fd < 0 )
 	{
@@ -1725,9 +1796,6 @@ do_cgi( void )
     char* binary;
     char* directory;
 
-    if ( method != METHOD_GET && method != METHOD_POST )
-	send_error( 501, "Not Implemented", "", "That method is not implemented for CGI." );
-
     /* If the socket happens to be using one of the stdin/stdout/stderr
     ** descriptors, move it to another descriptor so that the dup2 calls
     ** below don't screw things up.  We arbitrarily pick fd 3 - if there
@@ -1774,7 +1842,7 @@ do_cgi( void )
 	    /* Interposer process. */
 	    (void) close( p[0] );
 	    cgi_interpose_input( p[1] );
-	    exit( 0 );
+	    finish_request( 0 );
 	    }
 	(void) close( p[1] );
 	if ( p[0] != STDIN_FILENO )
@@ -1816,7 +1884,7 @@ do_cgi( void )
 	    /* Interposer process. */
 	    (void) close( p[1] );
 	    cgi_interpose_output( p[0], parse_headers );
-	    exit( 0 );
+	    finish_request( 0 );
 	    }
 	(void) close( p[0] );
 	if ( p[1] != STDOUT_FILENO )
@@ -1900,6 +1968,14 @@ cgi_interpose_input( int wfd )
     ssize_t r, r2;
     char buf[1024];
 
+    /* Set up the timeout for reading again, since we're in a sub-process. */
+#ifdef HAVE_SIGSET
+    (void) sigset( SIGALRM, handle_read_timeout );
+#else /* HAVE_SIGSET */
+    (void) signal( SIGALRM, handle_read_timeout );
+#endif /* HAVE_SIGSET */
+    (void) alarm( READ_TIMEOUT );
+
     c = request_len - request_idx;
     if ( c > 0 )
 	{
@@ -1929,6 +2005,7 @@ cgi_interpose_input( int wfd )
 	    break;
 	    }
 	c += r;
+	(void) alarm( READ_TIMEOUT );
 	}
     post_post_garbage_hack();
     }
@@ -1967,6 +2044,14 @@ cgi_interpose_output( int rfd, int parse_headers )
     ssize_t r, r2;
     char buf[1024];
 
+    /* Set up the timeout for writing again, since we're in a sub-process. */
+#ifdef HAVE_SIGSET
+    (void) sigset( SIGALRM, handle_write_timeout );
+#else /* HAVE_SIGSET */
+    (void) signal( SIGALRM, handle_write_timeout );
+#endif /* HAVE_SIGSET */
+    (void) alarm( WRITE_TIMEOUT );
+
     if ( ! parse_headers )
 	{
 	/* If we're not parsing headers, write out the default status line
@@ -1994,10 +2079,10 @@ cgi_interpose_output( int rfd, int parse_headers )
 
 	/* Slurp in all headers. */
 	headers_size = 0;
-	add_str( &headers, &headers_size, &headers_len, (char*) 0 );
+	add_data( &headers, &headers_size, &headers_len, (char*) 0, 0 );
 	for (;;)
 	    {
-	    r = read( rfd, buf, sizeof(buf) - 1 );
+	    r = read( rfd, buf, sizeof(buf) );
 	    if ( r < 0 && ( errno == EINTR || errno == EAGAIN ) )
 		{
 		sleep( 1 );
@@ -2008,15 +2093,14 @@ cgi_interpose_output( int rfd, int parse_headers )
 		br = &(headers[headers_len]);
 		break;
 		}
-	    buf[r] = '\0';
-	    add_str( &headers, &headers_size, &headers_len, buf );
+	    add_data( &headers, &headers_size, &headers_len, buf, r );
 	    if ( ( br = strstr( headers, "\015\012\015\012" ) ) != (char*) 0 ||
 		 ( br = strstr( headers, "\012\012" ) ) != (char*) 0 )
 		break;
 	    }
 
 	/* If there were no headers, bail. */
-	if ( headers[0] == '\0' )
+	if ( headers_len == 0 )
 	    return;
 
 	/* Figure out the status. */
@@ -2069,7 +2153,7 @@ cgi_interpose_output( int rfd, int parse_headers )
 	    continue;
 	    }
 	if ( r <= 0 )
-	    goto done;
+	    return;
 	for (;;)
 	    {
 	    r2 = my_write( buf, r );
@@ -2079,12 +2163,11 @@ cgi_interpose_output( int rfd, int parse_headers )
 		continue;
 		}
 	    if ( r2 != r )
-		goto done;
+		return;
 	    break;
 	    }
+	(void) alarm( WRITE_TIMEOUT );
 	}
-    done:
-    shutdown( conn_fd, SHUT_WR );
     }
 
 
@@ -2391,7 +2474,7 @@ send_error( int s, char* title, char* extra_header, char* text )
 #ifdef USE_SSL
     SSL_free( ssl );
 #endif /* USE_SSL */
-    exit( 1 );
+    finish_request( 1 );
     }
 
 
@@ -2576,9 +2659,9 @@ start_request( void )
     }
 
 static void
-add_to_request( char* str )
+add_to_request( char* str, size_t len )
     {
-    add_str( &request, &request_size, &request_len, str );
+    add_data( &request, &request_size, &request_len, str, len );
     }
 
 static char*
@@ -2704,6 +2787,7 @@ send_via_write( int fd, off_t size )
 		    return;
 		break;
 		}
+	    (void) alarm( WRITE_TIMEOUT );
 	    }
 	}
     }
@@ -2769,14 +2853,19 @@ my_write( void* buf, size_t size )
 
 
 #ifdef HAVE_SENDFILE
-static int
+static ssize_t
 my_sendfile( int fd, int s, off_t offset, size_t nbytes )
     {
 #ifdef HAVE_LINUX_SENDFILE
-	off_t lo = offset;
-	return sendfile( s, fd, &lo, nbytes );
+    off_t lo = offset;
+    return sendfile( s, fd, &lo, nbytes );
 #else /* HAVE_LINUX_SENDFILE */
-	return sendfile( fd, s, offset, nbytes, (struct sf_hdtr*) 0, (off_t*) 0, 0 );
+    int r;
+    r = sendfile( fd, s, offset, nbytes, (struct sf_hdtr*) 0, (off_t*) 0, 0 );
+    if ( r == 0 )
+	return nbytes;
+    else
+	return r;
 #endif /* HAVE_LINUX_SENDFILE */
     }
 #endif /* HAVE_SENDFILE */
@@ -2791,14 +2880,20 @@ add_str( char** bufP, size_t* bufsizeP, size_t* buflenP, char* str )
 	len = 0;
     else
 	len = strlen( str );
+    add_data( bufP, bufsizeP, buflenP, str, len );
+    }
 
+
+static void
+add_data( char** bufP, size_t* bufsizeP, size_t* buflenP, char* str, size_t len )
+    {
     if ( *bufsizeP == 0 )
 	{
 	*bufsizeP = len + 500;
 	*buflenP = 0;
 	*bufP = (char*) e_malloc( *bufsizeP );
 	}
-    else if ( *buflenP + len >= *bufsizeP )
+    else if ( *buflenP + len >= *bufsizeP )	/* allow for NUL */
 	{
 	*bufsizeP = *buflenP + len + 500;
 	*bufP = (char*) e_realloc( (void*) *bufP, *bufsizeP );
@@ -2997,6 +3092,9 @@ get_method_str( int m )
 	case METHOD_GET: return "GET";
 	case METHOD_HEAD: return "HEAD";
 	case METHOD_POST: return "POST";
+	case METHOD_PUT: return "PUT";
+	case METHOD_DELETE: return "DELETE";
+	case METHOD_TRACE: return "TRACE";
 	default: return "UNKNOWN";
 	}
     }
@@ -3262,7 +3360,7 @@ static void
 handle_write_timeout( int sig )
     {
     syslog( LOG_INFO, "%.80s connection timed out writing", ntoa( &client_addr ) );
-    exit( 1 );
+    finish_request( 1 );
     }
 
 
